import { IEventTag, IFormTag, ITrackFormTag } from '../types'
import { interceptRequests } from './request-interception'
import { getHost } from '../tracktor'

const spec: [IEventTag, IEventTag, IFormTag, ITrackFormTag, IEventTag] = [
  {
    framework: 'dom',
    url: 'http://localhost:4444/',
    taggingMethod: 'event',
    selector:
      'html > body > div#top > main:nth-child(3) > section#forms > form:nth-child(4) > fieldset#forms__textareas > p:nth-child(2) > textarea#textarea',
    handler: 'focus',
    event: 'Focused Textarea'
  },
  {
    framework: 'dom',
    url: 'http://localhost:4444/',
    taggingMethod: 'event',
    selector:
      'html > body > div#top > main:nth-child(3) > section#embedded > article#embedded__images > div:nth-child(2) > p:nth-child(2) > img:nth-child(1)',
    handler: 'click',
    event: 'Click Picture'
  },
  {
    framework: 'dom',
    url: 'http://localhost:4444',
    taggingMethod: 'form',
    formAction: 'identify',
    handler: 'submit',
    selector:
      'html > body > div#top > main:nth-child(3) > section#forms > div.log-form > form:nth-child(2)',
    bindings: [
      {
        selector: 'input[name="name"]',
        field: 'name',
        userIdentifier: false
      },
      {
        selector: 'input[name="email"]',
        field: 'email',
        userIdentifier: true
      }
    ]
  },
  {
    framework: 'dom',
    url: 'http://localhost:4444',
    taggingMethod: 'form',
    formAction: 'track',
    handler: 'submit',
    event: 'Submitted Form with Properties',
    selector:
      'html > body > div#top > main:nth-child(3) > section#forms > div.favorite-form > form:nth-child(2)',
    bindings: [
      {
        selector: 'input[name="food"]',
        field: 'food'
      },
      {
        selector: 'input[name="place"]',
        field: 'place'
      }
    ]
  },
  {
    framework: 'dom',
    url: 'http://localhost:4444/',
    taggingMethod: 'event',
    selector:
      'html > body > div#top > div.panel-footer > div.fusion-button-wrapper > a.fusion-button',
    handler: 'click',
    event: 'Clicked Link With Embedded Markup'
  }
]

// Map of all requests we want to stub out with local values
const responseMap = {
  [getHost()]: spec,
  'api.segment.io': { success: true }
}

describe('Tracktor', () => {
  beforeAll(async () => {
    interceptRequests(page, responseMap)
    await page.goto('https://localhost:4444')
    await page.waitForFunction('Tracktor.instance.isInitialized()')
  })

  it.skip('should emit correct track event on focus of <textarea> elements', async () => {
    const textareaFocusTag = spec[0]
    await expect(page).toClick(textareaFocusTag.selector)

    const ajsTrackCall = await page
      .waitForRequest(request =>
        request.url().startsWith('https://api.segment.io')
      )
      .then(e => e.postData())
      .then(JSON.parse)

    expect(ajsTrackCall.event).toEqual(textareaFocusTag.event)
  })

  it('should emit correct track event on click of an image', async () => {
    const imageClickedTag = spec[1]
    await expect(page).toClick(imageClickedTag.selector)

    const ajsTrackCall = await page
      .waitForRequest(request =>
        request.url().startsWith('https://api.segment.io/v1/t')
      )
      .then(e => e.postData())
      .then(JSON.parse)

    expect(ajsTrackCall.event).toEqual(imageClickedTag.event)
  })

  it('should emit an identify call when a tagged form (of formAction=identify) is submitted', async () => {
    const formSubmittedTag = spec[2]
    const name = 'John Doe'
    const email = 'jdoe@segment.com'

    // Fill out the form
    await expect(page).toFillForm(formSubmittedTag.selector, { name, email })

    // Click the button within the form
    await expect(page).toClick(formSubmittedTag.selector + ' > button')

    const ajsIdentifyCall = await page
      .waitForRequest(request =>
        request.url().startsWith('https://api.segment.io/v1/i')
      )
      .then(e => e.postData())
      .then(JSON.parse)

    expect(ajsIdentifyCall.type).toEqual('identify')
    expect(ajsIdentifyCall.traits).toEqual({ name })
    expect(ajsIdentifyCall.userId).toEqual(email)
    expect(ajsIdentifyCall.context.visual_tagger).toBeTruthy()
  })

  it('should emit an track call when a tagged form (of formAction=track) is submitted', async () => {
    const formSubmittedTag = spec[3]
    const food = 'Pizza'
    const place = 'Vancouver'

    // TODO: extract out tests that need to reset page context before runs as a separate block
    // (basically all form submissions)
    await page.goto('https://localhost:4444')

    // Fill out the form
    await expect(page).toFillForm(formSubmittedTag.selector, { food, place })

    // Click the button within the form
    await expect(page).toClick(formSubmittedTag.selector + ' > button')

    const ajsTrackCall = await page
      .waitForRequest(request =>
        request.url().startsWith('https://api.segment.io/v1/t')
      )
      .then(e => e.postData())
      .then(JSON.parse)

    expect(ajsTrackCall.type).toEqual('track')
    expect(ajsTrackCall.properties).toEqual({ place, food })
    expect(ajsTrackCall.context.visual_tagger).toBeTruthy()
  })

  it('should emit correct track event on click of an element with embedded markup', async () => {
    const embeddedMarkupClickTag = spec[4]
    // TODO: extract out tests that need to reset page context before runs as a separate block
    // (basically all form submissions)
    await page.goto('https://localhost:4444')

    await expect(page).toClick(
      embeddedMarkupClickTag.selector + ' > span.fusion-button-text'
    )

    const ajsTrackCall = await page
      .waitForRequest(request =>
        request.url().startsWith('https://api.segment.io')
      )
      .then(e => e.postData())
      .then(JSON.parse)

    expect(ajsTrackCall.event).toEqual(embeddedMarkupClickTag.event)
  })
})
