import {
  ITag,
  IFormTag,
  IEventTag,
  ITrackFormTag,
  ITrackTagBinding,
  IIdentifyTagBinding,
  SupportedDomEvents
} from './types'

const bubbleEvents = ['submit']

function isEventTag(tag: ITag): tag is IEventTag {
  return (
    (tag as IEventTag).taggingMethod === 'event' ||
    (tag as IEventTag).taggingMethod === undefined
  )
}

function isIdentifyFormTag(tag: ITag): tag is IFormTag {
  return (
    (tag as IFormTag).formAction === 'identify' &&
    (tag as IFormTag).taggingMethod === 'form'
  )
}

function isTrackFormTag(tag: ITag): tag is ITrackFormTag {
  return (
    (tag as IFormTag).formAction === 'track' &&
    (tag as IFormTag).taggingMethod === 'form'
  )
}

type EventToTagMap = { [k in SupportedDomEvents]: ITag[] }
type EventToListenerMap = { [k in SupportedDomEvents]: EventListener }

export const HOSTS = {
  staging: 'd3w1e142m55pwn.cloudfront.net',
  production: 'd1wnclalxop6x4.cloudfront.net'
}

export const getHost = () => {
  return HOSTS[process.env!.ENV!] || HOSTS['production']
}

const getBucketPath = (workspaceId: string, sourceId: string) => {
  const host = getHost()
  return `https://${host}/instrumentation-spec-${workspaceId}-${sourceId}.json`
}

export default class Tracktor {
  private initialized: boolean
  private listenerMap: EventToListenerMap
  static instance: Tracktor
  constructor(
    private workspaceId: string,
    private sourceId: string,
    private documentNode: Document
  ) {
    if (!this.workspaceId || !this.sourceId) {
      throw new Error('workspaceId and sourceId must be specified')
    }
    if (Tracktor.instance) {
      return Tracktor.instance
    }
    Tracktor.instance = this
  }

  public start(): Promise<void> {
    return Promise.resolve()
      .then(this.getInstrumentationSpec)
      .then(this.load)
  }

  public restart(specs: ITag[]): void {
    this.removeListeners()
    this.load(specs)
  }

  public isInitialized() {
    return this.initialized
  }

  private load = (specs: ITag[]) => {
    this.setListenerMap(this.toSpecMapByHandlerKey(specs))
    // TODO: Guard against this being invoked more than once
    this.initialized = true
    this.installListeners()
  }

  private getInstrumentationSpec = () => {
    return fetch(getBucketPath(this.workspaceId, this.sourceId), {
      headers: {
        'Cache-Control': 'no-cache'
      }
    }).then<ITag[]>(res => (res.ok ? res.json() : []))
  }

  private toSpecMapByHandlerKey(specs: ITag[]) {
    return specs.reduce(
      (res, spec) => ({
        ...res,
        [spec.handler]: [...(res[spec.handler] || []), spec]
      }),
      {} as EventToTagMap
    )
  }

  /**
   * Expands the "radius" of the selector by including ancestor elements up to a height defined
   * by maxAncestorHeight
   */
  private isMatch(
    tag: ITag,
    childElement: Element,
    maxAncestorHeight: number = 3
  ) {
    let curEl = childElement
    let curHeight = 0
    while (curEl && curHeight < maxAncestorHeight) {
      if (curEl.matches(tag.selector)) {
        return true
      }
      curEl = curEl.parentElement!
      curHeight++
    }
    return false
  }

  /**
   * For each event that a user has declared in their instrumentation spec,
   * this function will set an internal map of event names to handlers
   */
  private setListenerMap = (eventToTagMap: EventToTagMap) => {
    this.listenerMap = Object.entries(eventToTagMap).reduce(
      (listenerMap, [event, tags]) => {
        return {
          ...listenerMap,
          [event]: e => {
            const tag = tags.find(s => this.isMatch(s, e.target))

            if (!tag) {
              return
            }

            if (isEventTag(tag)) {
              this.handleEventTag(tag)
              return
            }

            if (isIdentifyFormTag(tag)) {
              this.handleIdentifyFormTag(tag)
              return
            }

            if (isTrackFormTag(tag)) {
              this.handleTrackFormTag(tag)
              return
            }
          }
        }
      },
      {} as EventToListenerMap
    )
  }

  private handleEventTag(tag: IEventTag) {
    window.analytics.track(
      tag.event,
      {},
      {
        context: {
          visual_tagger: true
        }
      }
    )
  }

  private handleIdentifyFormTag(tag: IFormTag) {
    const identifyBindings = tag.bindings as IIdentifyTagBinding[]
    const userIdentifier = identifyBindings.find(b => b.userIdentifier)
    const bindings = identifyBindings.filter(b => !b.userIdentifier)

    if (!userIdentifier) {
      console.error(
        `A unique identifier must be provided as a property binding for the form tagged at ${
          tag.selector
        } `
      )
      return
    }

    const formEl = this.documentNode.querySelector<HTMLFormElement>(
      tag.selector
    )

    if (!formEl) {
      console.error(
        `form tagged at ${tag.selector} cannot be found in the current page `
      )
      return
    }

    const identifierInput = formEl.querySelector<HTMLInputElement>(
      userIdentifier.selector
    )

    if (!identifierInput) {
      console.error(
        `Identifier input tagged at ${
          userIdentifier.selector
        } cannot be found in the current form`
      )
      return
    }

    const traits = bindings.reduce((traits, binding) => {
      const fieldValue = formEl.querySelector<HTMLInputElement>(
        binding.selector
      )

      return {
        ...traits,
        [binding.field]: fieldValue && fieldValue.value
      }
    }, {})

    window.analytics.identify(identifierInput.value, traits, {
      context: {
        visual_tagger: true
      }
    })
  }

  private handleTrackFormTag(tag: ITrackFormTag) {
    const bindings = tag.bindings as ITrackTagBinding[]

    const formEl = this.documentNode.querySelector<HTMLFormElement>(
      tag.selector
    )

    if (!formEl) {
      console.error(
        `form tagged at ${tag.selector} cannot be found in the current page `
      )
      return
    }

    const properties = bindings.reduce((props, binding) => {
      const fieldValue = formEl.querySelector<HTMLInputElement>(
        binding.selector
      )

      return {
        ...props,
        [binding.field]: fieldValue && fieldValue.value
      }
    }, {})

    window.analytics.track(tag.event, properties, {
      context: {
        visual_tagger: true
      }
    })
  }

  private installListeners() {
    Object.entries(this.listenerMap).forEach(([event, listener]) => {
      const useCapture = !bubbleEvents.includes(event)
      this.documentNode.addEventListener(event, listener, useCapture)
    })
  }

  private removeListeners() {
    Object.entries(this.listenerMap).forEach(([event, listener]) => {
      const useCapture = !bubbleEvents.includes(event)
      this.documentNode.removeEventListener(event, listener, useCapture)
    })
  }
}
